from sparksampling.error import BadParamError
from sparksampling.proto.sampling_service_pb2 import (
    CLUSTER_SAMPLING_METHOD,
    RANDOM_SAMPLING_METHOD,
    SIMPLE_SAMPLING_METHOD,
    STRATIFIED_SAMPLING_METHOD,
)
from sparksampling.sample.base_sampling import SparkBaseSamplingJob
from sparksampling.sample.cluster_sampling_imp import ClusterSamplingImp
from sparksampling.sample.random_sampling_imp import RandomSamplingImp
from sparksampling.sample.simple_sampling_imp import SimpleSamplingImp
from sparksampling.sample.stratified_sampling_imp import StratifiedSamplingImp


class SamplingFactory(object):
    method_map = {
        RANDOM_SAMPLING_METHOD: RandomSamplingImp,
        STRATIFIED_SAMPLING_METHOD: StratifiedSamplingImp,
        SIMPLE_SAMPLING_METHOD: SimpleSamplingImp,
        CLUSTER_SAMPLING_METHOD: ClusterSamplingImp,
    }

    @classmethod
    def get_sampling_imp(cls, sampling_method, sampling_conf) -> SparkBaseSamplingJob:
        return cls._get_imp_class(sampling_method)(
            **cls._get_sampling_conf(sampling_method, sampling_conf)
        )

    @classmethod
    def _get_imp_class(cls, sampling_method):
        imp_class = cls.method_map.get(sampling_method, None)
        if not imp_class:
            raise BadParamError(f"No matching sampling method: {sampling_method}")
        return imp_class

    @classmethod
    def _get_sampling_conf(cls, sampling_method, sampling_conf):
        return cls.method_map[sampling_method].get_init_conf(sampling_conf)



import json

import grpc
from sparksampling_client.config import load_config
from sparksampling_client.proto import sampling_service_pb2_grpc
from sparksampling_client.proto.sampling_service_pb2 import SamplingRequest


class Session:
    def __init__(self, host=None, port=None):
        config = load_config()
        self.host = host or config.get("host")
        self.port = port or config.get("port")

    def make_request(self, request):
        with grpc.insecure_channel(f"{self.host}:{self.port}") as channel:
            stub = sampling_service_pb2_grpc.SparkSamplingServiceStub(channel)
            response = stub.SamplingJob(request)
        return response

    def apply_file(self, json_file):
        with open(json_file) as f:
            request = SamplingRequest(**json.load(f))
        return self.make_request(request)



import logging
import os
import re
import signal
import time
from functools import partial, wraps

import requests
from google.protobuf.json_format import MessageToDict
from pyspark.sql import SparkSession

from sparksampling.error import SERVER_ERROR, CustomErrorWithCode

logger = logging.getLogger("sparksampling")


def hump2underline(hunp_str):
    p = re.compile(r"([a-z]|\d)([A-Z])")
    sub = re.sub(p, r"\1_\2", hunp_str).lower()
    return sub


def hump2underline_dict(data):
    keys = list(data.keys())
    for key in keys:
        underline_key = hump2underline(key)
        if type(data[key]) is dict:
            data[underline_key] = hump2underline_dict(data[key])
        data[underline_key] = data.pop(key)
    return data


def underline2hump(underline_str):
    sub = re.sub(r"(_\w)", lambda x: x.group(1)[1].upper(), underline_str)
    return sub


def throw_exception(func=None, *, request_type=None, response_type=None):
    if func is None:
        return partial(throw_exception, request_type=request_type, response_type=response_type)

    if not request_type:
        request_type = func.__annotations__["request"]

    if not response_type:
        response_type = func.__annotations__["return"]

    @wraps(func)
    def wrapper(self, request, context):
        data = MessageToDict(request, preserving_proto_field_name=True)
        parent_request = request_type(**data)
        start_time = time.time()
        try:
            return func(self, request, context)
        except CustomErrorWithCode as e:
            logger.exception(e)
            return response_type(
                **e.error_response(), data=response_type.ResponseData(parent_request=parent_request)
            )
        except Exception as e:
            logger.exception(e)
            return response_type(
                code=SERVER_ERROR,
                message=str(e),
                data=response_type.ResponseData(parent_request=parent_request),
            )
        finally:
            end_time = time.time()
            cls_logger = getattr(self, "logger", logger)
            cls_logger.debug(f"execute in {str(end_time - start_time).split('.')[0]} seconds")

    return wrapper


def async_throw_exception(func=None, *, request_type=None, response_type=None):
    if func is None:
        return partial(throw_exception, request_type=request_type, response_type=response_type)

    if not request_type:
        request_type = func.__annotations__["request"]

    if not response_type:
        response_type = func.__annotations__["return"]

    @wraps(func)
    async def wrapper(self, request, context):
        data = MessageToDict(request, preserving_proto_field_name=True)
        parent_request = request_type(**data)
        start_time = time.time()
        try:
            return await func(self, request, context)
        except CustomErrorWithCode as e:
            logger.exception(e)
            return response_type(
                **e.error_response(), data=response_type.ResponseData(parent_request=parent_request)
            )
        except Exception as e:
            logger.exception(e)
            return response_type(
                code=SERVER_ERROR,
                message=str(e),
                data=response_type.ResponseData(parent_request=parent_request),
            )
        finally:
            end_time = time.time()
            cls_logger = getattr(self, "logger", logger)
            cls_logger.debug(f"execute in {str(end_time - start_time).split('.')[0]} seconds")

    return wrapper


def check_spark_session(func):
    # Check the spark driver status and exit if it is not responding
    # Server will be redeployed under K8S deployment
    def _check_spark_ui(logger, spark: SparkSession):
        try:
            if not spark.conf.get("spark.submit.deployMode") == "client":
                return True
        except Exception as e:
            logger.exception(e)
            logger.warning(f"can't access spark.conf")
            return False

        ui_port = spark.conf.get("spark.ui.port")
        url = f"http://localhost:{ui_port}"
        logger.debug(f"Check Spark is alive: {url}")
        try:
            response = requests.get(url)
        except Exception as e:
            logger.exception(e)
            logger.warning(f"Can't access Spark ui: {url}")
            return False
        return response.status_code == 200

    def exit_if_spark_dead(self):
        if not _check_spark_ui(self.logger, self.spark):
            cls_logger = getattr(self, "logger") or logger
            cls_logger.critical("Spark is dead, exiting...")
            os.kill(os.getpid(), signal.SIGINT)

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            r = func(self, *args, **kwargs)
        finally:
            exit_if_spark_dead(self)
        return r

    return wrapper



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sampling_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16sampling_service.proto"\xea\x01\n\x0fSamplingRequest\x12,\n\x0fsampling_method\x18\x01 \x01(\x0e\x32\x13.SamplingMethodType\x12$\n\x0b\x66ile_format\x18\x02 \x01(\x0e\x32\x0f.FileFormatType\x12$\n\rsampling_conf\x18\x03 \x01(\x0b\x32\r.SamplingConf\x12$\n\x0b\x66ormat_conf\x18\x04 \x01(\x0b\x32\x0f.FileFormatConf\x12\x12\n\ninput_path\x18\x05 \x01(\t\x12\x13\n\x0boutput_path\x18\x06 \x01(\t\x12\x0e\n\x06job_id\x18\x07 \x01(\t"\x1f\n\rCancelRequest\x12\x0e\n\x06job_id\x18\x01 \x01(\t"/\n\x0e\x43\x61ncelResponse\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t"\xcf\x01\n\x10SamplingResponse\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12,\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x1e.SamplingResponse.ResponseData\x1an\n\x0cResponseData\x12(\n\x0eparent_request\x18\x01 \x01(\x0b\x32\x10.SamplingRequest\x12\x14\n\x0csampled_path\x18\x02 \x01(\t\x12\x1e\n\x08hook_msg\x18\x03 \x03(\x0b\x32\x0c.HookMessage"#\n\x04\x64ict\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t"\xbe\x01\n\x0cSamplingConf\x12\x10\n\x08\x66raction\x18\x01 \x01(\t\x12\x18\n\x10with_replacement\x18\x02 \x01(\x08\x12\x0c\n\x04seed\x18\x03 \x01(\r\x12\x16\n\x0estratified_key\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\r\x12\x14\n\x0csampling_col\x18\x06 \x03(\t\x12\x10\n\x08group_by\x18\x07 \x01(\t\x12\x11\n\tgroup_num\x18\x08 \x01(\r\x12\x12\n\nensure_col\x18\t \x01(\x08"2\n\x0e\x46ileFormatConf\x12\x13\n\x0bwith_header\x18\x01 \x01(\x08\x12\x0b\n\x03sep\x18\x02 \x01(\t"\xb0\x08\n\x17RelationSamplingRequest\x12?\n\x0fsampling_stages\x18\x01 \x03(\x0b\x32&.RelationSamplingRequest.SamplingStage\x12?\n\x0frelation_stages\x18\x02 \x03(\x0b\x32&.RelationSamplingRequest.RelationStage\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12\x34\n\x17\x64\x65\x66\x61ult_sampling_method\x18\x04 \x01(\x0e\x32\x13.SamplingMethodType\x12,\n\x15\x64\x65\x66\x61ult_sampling_conf\x18\x05 \x01(\x0b\x32\r.SamplingConf\x12,\n\x13\x64\x65\x66\x61ult_file_format\x18\x06 \x01(\x0e\x32\x0f.FileFormatType\x12,\n\x13\x64\x65\x66\x61ult_format_conf\x18\x07 \x01(\x0b\x32\x0f.FileFormatConf\x12\x0f\n\x07\x64ry_run\x18\x08 \x01(\x08\x1a\x8e\x02\n\rSamplingStage\x12\x12\n\ninput_path\x18\x01 \x01(\t\x12\x12\n\ninput_name\x18\x02 \x01(\t\x12\x13\n\x0boutput_path\x18\x03 \x01(\t\x12\x12\n\noutput_col\x18\x04 \x03(\t\x12\x0c\n\x04name\x18\x05 \x01(\t\x12,\n\x0fsampling_method\x18\x06 \x01(\x0e\x32\x13.SamplingMethodType\x12$\n\rsampling_conf\x18\x07 \x01(\x0b\x32\r.SamplingConf\x12$\n\x0b\x66ile_format\x18\x08 \x01(\x0e\x32\x0f.FileFormatType\x12$\n\x0b\x66ormat_conf\x18\t \x01(\x0b\x32\x0f.FileFormatConf\x1a\xad\x01\n\x08Relation\x12\x15\n\rrelation_path\x18\x01 \x01(\t\x12\x15\n\rrelation_name\x18\x02 \x01(\t\x12\x14\n\x0crelation_col\x18\x03 \x01(\t\x12\x11\n\tinput_col\x18\x04 \x01(\t\x12$\n\x0b\x66ile_format\x18\x05 \x01(\x0e\x32\x0f.FileFormatType\x12$\n\x0b\x66ormat_conf\x18\x06 \x01(\x0b\x32\x0f.FileFormatConf\x1a\xf0\x01\n\rRelationStage\x12\x12\n\ninput_path\x18\x01 \x01(\t\x12\x12\n\ninput_name\x18\x02 \x01(\t\x12\x13\n\x0boutput_path\x18\x03 \x01(\t\x12\x12\n\noutput_col\x18\x04 \x03(\t\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x34\n\trelations\x18\x06 \x03(\x0b\x32!.RelationSamplingRequest.Relation\x12$\n\x0b\x66ile_format\x18\x07 \x01(\x0e\x32\x0f.FileFormatType\x12$\n\x0b\x66ormat_conf\x18\x08 \x01(\x0b\x32\x0f.FileFormatConf"\xbe\x02\n\x18RelationSamplingResponse\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x34\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32&.RelationSamplingResponse.ResponseData\x1aR\n\x0cSampleResult\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0csampled_path\x18\x02 \x01(\t\x12\x1e\n\x08hook_msg\x18\x03 \x03(\x0b\x32\x0c.HookMessage\x1ay\n\x0cResponseData\x12\x30\n\x0eparent_request\x18\x01 \x01(\x0b\x32\x18.RelationSamplingRequest\x12\x37\n\x07results\x18\x02 \x03(\x0b\x32&.RelationSamplingResponse.SampleResult"E\n\x0bHookMessage\x12\x11\n\thook_name\x18\x01 \x01(\t\x12\x0e\n\x06period\x18\x02 \x01(\t\x12\x13\n\x04meta\x18\x03 \x03(\x0b\x32\x05.pair""\n\x04pair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t*\x9d\x01\n\x12SamplingMethodType\x12\x12\n\x0eUNKNOWN_METHOD\x10\x00\x12\x1a\n\x16RANDOM_SAMPLING_METHOD\x10\x01\x12\x1e\n\x1aSTRATIFIED_SAMPLING_METHOD\x10\x02\x12\x1a\n\x16SIMPLE_SAMPLING_METHOD\x10\x03\x12\x1b\n\x17\x43LUSTER_SAMPLING_METHOD\x10\x04*9\n\x0e\x46ileFormatType\x12\x12\n\x0eUNKNOWN_FORMAT\x10\x00\x12\x13\n\x0f\x46ILE_FORMAT_CSV\x10\x01\x32\xca\x01\n\x14SparkSamplingService\x12\x34\n\x0bSamplingJob\x12\x10.SamplingRequest\x1a\x11.SamplingResponse"\x00\x12L\n\x13RelationSamplingJob\x12\x18.RelationSamplingRequest\x1a\x19.RelationSamplingResponse"\x00\x12.\n\tCancelJob\x12\x0e.CancelRequest\x1a\x0f.CancelResponse"\x00\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "sampling_service_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SAMPLINGMETHODTYPE._serialized_start = 2341
    _SAMPLINGMETHODTYPE._serialized_end = 2498
    _FILEFORMATTYPE._serialized_start = 2500
    _FILEFORMATTYPE._serialized_end = 2557
    _SAMPLINGREQUEST._serialized_start = 27
    _SAMPLINGREQUEST._serialized_end = 261
    _CANCELREQUEST._serialized_start = 263
    _CANCELREQUEST._serialized_end = 294
    _CANCELRESPONSE._serialized_start = 296
    _CANCELRESPONSE._serialized_end = 343
    _SAMPLINGRESPONSE._serialized_start = 346
    _SAMPLINGRESPONSE._serialized_end = 553
    _SAMPLINGRESPONSE_RESPONSEDATA._serialized_start = 443
    _SAMPLINGRESPONSE_RESPONSEDATA._serialized_end = 553
    _DICT._serialized_start = 555
    _DICT._serialized_end = 590
    _SAMPLINGCONF._serialized_start = 593
    _SAMPLINGCONF._serialized_end = 783
    _FILEFORMATCONF._serialized_start = 785
    _FILEFORMATCONF._serialized_end = 835
    _RELATIONSAMPLINGREQUEST._serialized_start = 838
    _RELATIONSAMPLINGREQUEST._serialized_end = 1910
    _RELATIONSAMPLINGREQUEST_SAMPLINGSTAGE._serialized_start = 1221
    _RELATIONSAMPLINGREQUEST_SAMPLINGSTAGE._serialized_end = 1491
    _RELATIONSAMPLINGREQUEST_RELATION._serialized_start = 1494
    _RELATIONSAMPLINGREQUEST_RELATION._serialized_end = 1667
    _RELATIONSAMPLINGREQUEST_RELATIONSTAGE._serialized_start = 1670
    _RELATIONSAMPLINGREQUEST_RELATIONSTAGE._serialized_end = 1910
    _RELATIONSAMPLINGRESPONSE._serialized_start = 1913
    _RELATIONSAMPLINGRESPONSE._serialized_end = 2231
    _RELATIONSAMPLINGRESPONSE_SAMPLERESULT._serialized_start = 2026
    _RELATIONSAMPLINGRESPONSE_SAMPLERESULT._serialized_end = 2108
    _RELATIONSAMPLINGRESPONSE_RESPONSEDATA._serialized_start = 2110
    _RELATIONSAMPLINGRESPONSE_RESPONSEDATA._serialized_end = 2231
    _HOOKMESSAGE._serialized_start = 2233
    _HOOKMESSAGE._serialized_end = 2302
    _PAIR._serialized_start = 2304
    _PAIR._serialized_end = 2338
    _SPARKSAMPLINGSERVICE._serialized_start = 2560
    _SPARKSAMPLINGSERVICE._serialized_end = 2762
# @@protoc_insertion_point(module_scope)



from sparksampling.engine.dummy_engine import DummyEngine
from sparksampling.evaluation.base_hook import BaseEvaluationHook


class DummyExtensionHook(BaseEvaluationHook):
    pass


DummyExtensionHook.register_pre_hook(DummyEngine)



import pytest

from sparksampling.proto.sampling_service_pb2 import (
    FILE_FORMAT_CSV,
    RANDOM_SAMPLING_METHOD,
    FileFormatConf,
    SamplingConf,
    SamplingRequest,
)


@pytest.mark.xfail
def test_random_sampling(grpc_stub):
    sampling_conf = SamplingConf(
        fraction="0.2",
        with_replacement=False,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    request = SamplingRequest(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path="s3a://datasampling-test/object",
        output_path="s3a://datasampling-test/object-unittest",
        job_id="test_random_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")



from sparksampling.engine_factory import EngineFactory
from sparksampling.mixin import LogMixin
from sparksampling.proto import sampling_service_pb2_grpc
from sparksampling.proto.sampling_service_pb2 import (
    CancelRequest,
    CancelResponse,
    RelationSamplingRequest,
    RelationSamplingResponse,
    SamplingRequest,
    SamplingResponse,
)
from sparksampling.utilities import throw_exception


class GRPCService(sampling_service_pb2_grpc.SparkSamplingServiceServicer, LogMixin):
    def __init__(self, parent):
        self.parent = parent

    @staticmethod
    def register_engine(parent):
        EngineFactory.register_all_engine()

    @staticmethod
    def get_worker_num():
        return EngineFactory.get_engine_total_worker()

    @staticmethod
    def add_to_server(parent, server):
        """Register the service to the GRPC Server
        :param parent: application instance
        :param server: server instance
        :return:
        """
        GRPCService.logger.info("Add GRPCService Service to server")
        sampling_service_pb2_grpc.add_SparkSamplingServiceServicer_to_server(
            GRPCService(parent), server
        )

    @throw_exception
    def SamplingJob(self, request: SamplingRequest, context) -> SamplingResponse:
        data = EngineFactory.message_to_dict(request)
        parent_request = SamplingRequest(**data)
        engine = EngineFactory.get_engine(self.parent, SamplingRequest, **data)
        output_path, hook_msg = engine.submit()
        return SamplingResponse(
            code=0,
            message="",
            data=SamplingResponse.ResponseData(
                parent_request=parent_request, sampled_path=output_path, hook_msg=hook_msg
            ),
        )

    @throw_exception
    def RelationSamplingJob(
        self, request: RelationSamplingRequest, context
    ) -> RelationSamplingResponse:
        data = EngineFactory.message_to_dict(request)
        parent_request = RelationSamplingRequest(**data)
        engine = EngineFactory.get_engine(self.parent, RelationSamplingRequest, **data)
        results = engine.submit()

        return RelationSamplingResponse(
            code=0,
            message="",
            data=RelationSamplingResponse.ResponseData(
                parent_request=parent_request, results=results
            ),
        )

    def CancelJob(self, request: CancelRequest, context) -> CancelResponse:
        try:
            EngineFactory.cancel_job(self.parent, request.job_id)
        except Exception as e:
            self.logger.exception(e)
            return CancelResponse(code=5000, message=str(e))
        return CancelResponse(code=0, message="ok")



from concurrent.futures import ThreadPoolExecutor
from graphlib import CycleError, TopologicalSorter
from pprint import pformat
from typing import Dict, List

from sparksampling.engine.base_engine import SparkBaseEngine
from sparksampling.engine_factory import EngineFactory
from sparksampling.error import BadParamError
from sparksampling.file_format import FileFormatFactory
from sparksampling.mixin import LogMixin, acquire_worker, record_job_id
from sparksampling.proto.sampling_service_pb2 import RelationSamplingRequest
from sparksampling.sample import SamplingFactory
from sparksampling.utilities import check_spark_session


def _check_param(conf):
    def check_duplicated_name(name):
        if name not in name_set:
            name_set.add(name)
        else:
            raise BadParamError(f"Duplicated name: {name}")

    name_set = set()
    for sampling_stage in conf.get("sampling_stages", []):
        check_duplicated_name(sampling_stage["name"])
        if not (sampling_stage.get("input_path") or sampling_stage.get("input_name")):
            raise BadParamError(
                f"Need at least one sampling source(input_path or input_name) for relation {pformat(sampling_stage)}"
            )

    for relation_stage in conf.get("relation_stage", []):
        # check duplicate name
        check_duplicated_name(relation_stage["name"])
        for relation in relation_stage.get("relations", []):
            if not (relation.get("relation_path") or relation.get("relation_name")):
                raise BadParamError(
                    f"Need at least one relation source(relation_path or relation_name) for relation {pformat(relation_stage)}"
                )
            if not (relation.get("relation_col") or relation.get("input_col")):
                raise BadParamError(
                    f"Need at least one col(relation_col or input_col) for relation {pformat(relation_stage)}"
                )


def _set_default(conf):
    # allow empty stage for export or sampling only job
    conf.setdefault("sampling_stages", [])
    conf.setdefault("relation_stages", [])

    default_sampling_method = conf.get("default_sampling_method")
    default_sampling_conf = conf.get("default_sampling_conf")
    default_file_format = conf.get("default_file_format")
    default_format_conf = conf.get("default_format_conf")
    for sampling_stage in conf["sampling_stages"]:
        sampling_stage.setdefault("sampling_method", default_sampling_method)
        sampling_stage.setdefault("sampling_conf", default_sampling_conf)
        sampling_stage.setdefault("file_format", default_file_format)
        sampling_stage.setdefault("format_conf", default_format_conf)
    for relation_stage in conf["relation_stages"]:
        relation_stage.setdefault("file_format", default_file_format)
        relation_stage.setdefault("format_conf", default_format_conf)
        for relation in relation_stage.get("relations", []):
            if not relation.get("relation_col"):
                relation["relation_col"] = relation["input_col"]
            if not relation.get("input_col"):
                relation["input_col"] = relation["relation_col"]
            relation.setdefault("file_format", default_file_format)
            relation.setdefault("format_conf", default_format_conf)
    return conf


class JobStage(LogMixin):
    def __init__(self, spark, stage_config, name_df_map):
        self.spark = spark
        self.stage_config = stage_config
        self.name_df_map = name_df_map

        self._df = None
        self.sampled_path = None
        self.file_format_imp = None
        self.sample_imp = None
        self.pre_metas = []
        self.post_metas = []

        self._init_sample_imp()
        self._init_file_format()

    def _init_file_format(self):
        self.file_format_imp = FileFormatFactory.get_file_imp(
            self.spark, self.file_format, self.format_conf
        )

    def _init_sample_imp(self):
        if not (self.sampling_method and self.sampling_conf):
            return
        self.sample_imp = SamplingFactory.get_sampling_imp(self.sampling_method, self.sampling_conf)

    def _build_relation_dataframe(self, input_df, relation):
        relation_name = relation.get("relation_name")
        if relation_name:
            relation_df = self.name_df_map[relation_name]
        else:
            # using relation_path to construct dataframe
            relation_file_format = FileFormatFactory.get_file_imp(
                self.spark, relation["file_format"], relation["format_conf"]
            )
            # no need to worry if spark read the same csv twice, spark only read once
            relation_df = relation_file_format.read(relation["relation_path"])

        # python's way for self._df.join(relation_df, self._df.input_col == relation_df.relation_col, 'semi')
        cond = [
            getattr(input_df, relation["input_col"])
            == getattr(relation_df, relation["relation_col"])
        ]
        return input_df.join(relation_df, cond, how="semi")

    def build_dataframe(self, pre_hook=None, post_hook=None):
        self.log.debug(f"Starting building stage {self.name}")
        if self.input_name:
            self._df = self.name_df_map[self.input_name]
        else:
            self._df = self.file_format_imp.read(self.input_path)

        if pre_hook:
            self._df, self.pre_metas = pre_hook(self._df)

        if self.sample_imp:
            self._df = self.sample_imp.run(self._df)

        for relation in self.relations:
            self._df = self._build_relation_dataframe(self._df, relation)

        if post_hook:
            self._df, self.post_metas = post_hook(self._df)

        self.name_df_map[self.name] = self._df

    @property
    def relations(self):
        return self.stage_config.get("relations", [])

    @property
    def name(self):
        return self.stage_config.get("name")

    @property
    def sampling_method(self):
        return self.stage_config.get("sampling_method")

    @property
    def sampling_conf(self):
        return self.stage_config.get("sampling_conf")

    @property
    def file_format(self):
        return self.stage_config.get("file_format")

    @property
    def format_conf(self):
        return self.stage_config.get("format_conf")

    @property
    def output_path(self):
        return self.stage_config.get("output_path")

    @property
    def output_col(self):
        return self.stage_config.get("output_col")

    @property
    def input_name(self):
        return self.stage_config.get("input_name")

    @property
    def input_path(self):
        return self.stage_config.get("input_path")

    def export_dataframe(self):
        if self.output_path:
            self.log.debug(f"Exporting stage {self.name} to {self.output_path}")
            self.sampled_path = self.file_format_imp.write(
                self._df, self.output_path, self.output_col
            )
        else:
            self.log.debug(f"Nothing to export for {self.name}")


class RMSEngine(SparkBaseEngine):
    # Relationship Mapping Sampling
    def __init__(
        self,
        parent,
        sampling_stages: List[Dict],
        relation_stages: List[Dict],
        job_id: str,
        dry_run: bool,
    ):
        super(RMSEngine, self).__init__()
        self.parent = parent
        self.sampling_stages = sampling_stages
        self.relation_stages = relation_stages
        self.job_id = job_id
        self.dry_run = dry_run

        self.name_stage_map = dict()
        self._init_name_stage_map()

    def _init_name_stage_map(self):
        for sampling_stage in self.sampling_stages:
            self.name_stage_map[sampling_stage["name"]] = sampling_stage
            self.name_stage_map["type"] = "sample"
        for relation_stages in self.relation_stages:
            self.name_stage_map[relation_stages["name"]] = relation_stages
            self.name_stage_map["type"] = "relation"

    @acquire_worker
    @check_spark_session
    @record_job_id
    def submit(self, *args, **kwargs):
        try:
            build_order = self._get_stage_order()
        except CycleError as e:
            raise BadParamError(f"Loop detected: {str(e)}")

        spark_job_seq = self._process_spark_seq(build_order)
        result = self._get_seq_result(spark_job_seq)
        return result

    def _get_stage_order(self):
        topological_map = dict()

        for sampling_stage in self.sampling_stages:
            rely_set = topological_map.setdefault(sampling_stage["name"], set())
            rely_name = sampling_stage.get("input_name")
            if rely_name:
                rely_set.add(rely_name)

        for relation_stage in self.relation_stages:
            rely_set = topological_map.setdefault(relation_stage["name"], set())
            rely_name = relation_stage.get("input_name")
            if rely_name:
                rely_set.add(rely_name)
            for relation in relation_stage.get("relations", []):
                relation_name = relation.get("relation_name")
                if relation_name:
                    rely_set.add(relation_name)
        self.log.debug(f"topological_map is {pformat(topological_map)}")
        build_order = list(TopologicalSorter(topological_map).static_order())
        self.log.debug(f"topological sort result: {build_order}")
        return build_order

    def _process_spark_seq(self, build_order):
        job_seq = []
        name_df_map = dict()
        for name in build_order:
            stage = JobStage(self.spark, self.name_stage_map[name], name_df_map)
            stage.build_dataframe(self.pre_hook, self.post_hook)
            job_seq.append(stage)
        if self.dry_run:
            self.log.info("Dry run, skip export dataframe")
            return
        self.log.info("Generated spark job sequence, start exporting dataframe...")

        executors = int(self.spark.conf.get("spark.executor.instances", "1"))
        with ThreadPoolExecutor(max_workers=executors) as executor:
            for stage in job_seq:
                executor.submit(stage.export_dataframe)
        return job_seq

    def _get_seq_result(self, seq):
        self.log.debug("Start collecting result...")
        if self.dry_run:
            self.log.info("Dry run, nothing output")
            return []
        result = []
        for job in seq:
            if job.sampled_path:
                result.append(
                    {
                        "name": job.name,
                        "sampled_path": job.sampled_path,
                        "hook_msg": job.pre_metas + job.post_metas,
                    }
                )
        self.log.info(f"result collected: \n{pformat(result)}")
        return result

    @classmethod
    def config(cls, kwargs):
        _check_param(kwargs)
        try:
            kwargs = _set_default(kwargs)
            sampling_stages = kwargs.pop("sampling_stages")
            relation_stages = kwargs.pop("relation_stages")
            job_id = kwargs.pop("job_id")
            dry_run = kwargs.pop("dry_run", False)
        except KeyError as e:
            cls.log.info(f"Missing required parameters {str(e)}")
            raise BadParamError(f"Missing required parameters {str(e)}")

        config_dict = {
            "sampling_stages": sampling_stages,
            "relation_stages": relation_stages,
            "job_id": job_id,
            "dry_run": dry_run,
        }
        cls.log.info(f"Initializing job conf... \n {pformat(config_dict)}")
        return config_dict

    @classmethod
    def is_matching(cls, request_type):
        return request_type == RelationSamplingRequest


EngineFactory.register(RMSEngine)



# import all .py from in this dir
import glob
import importlib
from os.path import basename, dirname, isfile, join

modules = glob.glob(join(dirname(__file__), "*.py"))
sub_packages = (basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py"))
packages = (str(__package__) + "." + i for i in sub_packages)
[importlib.import_module(p) for p in packages]

importlib.import_module("sparksampling.evaluation")



from pyspark.sql import DataFrame

from sparksampling.evaluation.base_hook import BaseEvaluationHook


class EmptyDetectHook(BaseEvaluationHook):
    def _process(self, df: DataFrame):
        msg = {"is_not_empty": bool(df.take(1))}

        return df, msg


EmptyDetectHook.register_pre_hook_all()
EmptyDetectHook.register_post_hook_all()



import threading


class JobThread(threading.Thread):
    def __init__(self, target, args=None, kwargs=None):
        super(JobThread, self).__init__()
        self.target = target
        if not args:
            args = []
        self.args = args
        if not kwargs:
            kwargs = dict()
        self.kwargs = kwargs
        self.result = None

    def run(self):
        self.result = self.target(*self.args, **self.kwargs)

    def get_result(self):
        return self.result



from pyspark.sql import DataFrame

from sparksampling.mixin import LogMixin, SparkMixin


class SparkBaseSamplingJob(SparkMixin, LogMixin):
    cls_args = []

    def __init__(self, *args, **kwargs):
        ...

    @classmethod
    def get_init_conf(cls, conf):
        if not cls.cls_args:
            cls.log.warning("No parameters to initialize ?")
            return dict
        init_conf = dict()
        for arg in cls.cls_args:
            if conf.get(arg):
                init_conf[arg] = conf.get(arg)
        return init_conf

    def run(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError



BAD_PARAM_ERROR = 4000
PROCESS_ERROR = 4004
PROCESS_PRE_HOOK_ERROR = 4011
PROCESS_POST_HOOK_ERROR = 4012
EXHAUSTED_ERROR = 4006
SERVER_ERROR = 5000


class CustomErrorWithCode(Exception):
    def __init__(self, code, error_info):
        super().__init__(self)
        self.code = code
        self.errorinfo = error_info

    def __str__(self):
        return self.errorinfo

    def error_response(self):
        return {
            "code": self.code,
            "message": self.errorinfo,
        }


class ExhaustedError(CustomErrorWithCode):
    def __init__(self, msg: str):
        super(ExhaustedError, self).__init__(EXHAUSTED_ERROR, msg)


class BadParamError(CustomErrorWithCode):
    def __init__(self, msg: str):
        super(BadParamError, self).__init__(BAD_PARAM_ERROR, msg)


class ProcessError(CustomErrorWithCode):
    def __init__(self, msg: str):
        super(ProcessError, self).__init__(PROCESS_ERROR, msg)


class ProcessPreHookError(CustomErrorWithCode):
    def __init__(self, msg: str):
        super(ProcessPreHookError, self).__init__(PROCESS_PRE_HOOK_ERROR, msg)


class ProcessPostHookError(CustomErrorWithCode):
    def __init__(self, msg: str):
        super(ProcessPostHookError, self).__init__(PROCESS_POST_HOOK_ERROR, msg)



class HookMsg(object):
    def __init__(self, hook, msg):
        self.hook_name = hook.__name__ if isinstance(hook, type) else hook.__class__.__name__
        self.msg = msg

    def generate_proto_msg(self, period):
        proto_pair_msg = []
        for k, v in self.msg.items():
            proto_pair_msg.append({"key": str(k), "value": str(v)})

        return {
            "hook_name": self.hook_name,
            "period": period,
            "meta": proto_pair_msg,
        }



from typing import Dict

from pyspark.sql import DataFrame

from sparksampling.engine_factory import EngineFactory
from sparksampling.evaluation.hook_msg import HookMsg


class BaseEvaluationHook:
    def process(self, df: DataFrame) -> (DataFrame, HookMsg):
        df, msg = self._process(df)
        return df, HookMsg(self, msg)

    def _process(self, df) -> (DataFrame, Dict):
        raise NotImplementedError

    @classmethod
    def register_pre_hook(cls, engine):
        engine.register_pre_hook(cls)

    @classmethod
    def register_post_hook(cls, engine):
        engine.register_post_hook(cls)

    @classmethod
    def _get_all_engine(cls):
        return EngineFactory.engine_cls

    @classmethod
    def register_pre_hook_all(cls):
        for engine in cls._get_all_engine():
            cls.register_pre_hook(engine)

    @classmethod
    def register_post_hook_all(cls):
        for engine in cls._get_all_engine():
            cls.register_post_hook(engine)


class ByPassEvaluationHook(BaseEvaluationHook):
    # this hook for test use
    def _process(self, df):
        return df, {"msg": "bypass"}


# ByPassEvaluationHook.register_pre_hook_all()
# ByPassEvaluationHook.register_post_hook_all()



import random

from sparksampling.sample.base_sampling import SparkBaseSamplingJob


class SimpleSamplingImp(SparkBaseSamplingJob):
    cls_args = ["fraction", "with_replacement", "seed", "count", "sampling_col"]

    def __init__(self, *args, **kwargs):
        super(SimpleSamplingImp, self).__init__(*args, **kwargs)
        self.fraction = float(kwargs.pop("fraction", 1))
        self.with_replacement = kwargs.pop("with_replacement", False)
        self.seed = kwargs.pop("seed", random.randint(1, 65535))
        self.count = kwargs.pop("count")
        self.sampling_col = list(kwargs.pop("sampling_col", []))

    def run(self, df):
        if self.sampling_col:
            df = df[self.sampling_col]
        return df.sample(
            fraction=self.fraction, seed=self.seed, withReplacement=self.with_replacement
        ).limit(self.count)



import importlib
from pprint import pformat

from google.protobuf.json_format import MessageToDict

from sparksampling.error import BadParamError, ProcessError
from sparksampling.mixin import LogMixin


class EngineFactory(LogMixin):
    engine_cls = set()

    @classmethod
    def choose_engine(cls, request_type):
        engine_cls = None
        for engine in cls.engine_cls:
            if engine.is_matching(request_type):
                if engine_cls:
                    raise ProcessError(
                        f"Duplicated Engine Found: {engine.__name__} and {engine_cls.__name__}"
                    )
                engine_cls = engine
        return engine_cls

    @classmethod
    def get_engine(cls, parent, request_type, **kwargs):
        engine_cls = cls.choose_engine(request_type)
        if not engine_cls:
            raise BadParamError(f"No matching engine for this job, \n{pformat(kwargs, indent=2)}")
        cls.log.info(f"Using {engine_cls.__name__}")
        return engine_cls(parent, **engine_cls.config(kwargs))

    @classmethod
    def message_to_dict(cls, message):
        data = MessageToDict(message, use_integers_for_enums=True, preserving_proto_field_name=True)
        return data

    @classmethod
    def cancel_job(cls, parent, job_id):
        for engine in cls.engine_cls:
            engine.stop(parent, job_id)

    @classmethod
    def register(cls, engine_class):
        if engine_class in cls.engine_cls:
            return
        cls.log.info(
            f"Register engine: {engine_class.__name__}, allocate {engine_class.guarantee_worker} workers"
        )
        cls.engine_cls.add(engine_class)

    @staticmethod
    def register_all_engine():
        importlib.import_module("sparksampling.engine")

    @classmethod
    def get_engine_total_worker(cls):
        return sum(engine.guarantee_worker for engine in cls.engine_cls)



import os
from functools import reduce

import pandas
import pytest
from utilities import JobThread

from sparksampling.error import BAD_PARAM_ERROR, PROCESS_ERROR
from sparksampling.proto.sampling_service_pb2 import (
    CLUSTER_SAMPLING_METHOD,
    FILE_FORMAT_CSV,
    RANDOM_SAMPLING_METHOD,
    SIMPLE_SAMPLING_METHOD,
    STRATIFIED_SAMPLING_METHOD,
    UNKNOWN_FORMAT,
    UNKNOWN_METHOD,
    CancelRequest,
    FileFormatConf,
    SamplingConf,
    SamplingRequest,
)

dir_prefix = os.path.abspath(os.path.dirname(__file__))
input_path = os.path.join(dir_prefix, "../data/10w_x_10.csv")


def test_random_sampling_fraction(grpc_stub):
    sampling_col = ["# id", "X_0", "X_1", "X_2", "y"]
    sampling_conf = SamplingConf(fraction="0.2", with_replacement=False, sampling_col=sampling_col)
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/test_random_sampling_fraction")
    request = SamplingRequest(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_random_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")
    if os.path.exists(response.data.sampled_path):
        assert sorted(list(pandas.read_csv(response.data.sampled_path).columns)) == sorted(
            list(sampling_col)
        )


def test_random_sampling_count(grpc_stub):
    sampling_col = ["# id", "X_0", "X_1", "X_2", "y"]
    sampling_conf = SamplingConf(count=200, with_replacement=False, sampling_col=sampling_col)
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/test_random_sampling_count")
    request = SamplingRequest(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_random_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")
    if os.path.exists(response.data.sampled_path):
        assert sorted(list(pandas.read_csv(response.data.sampled_path).columns)) == sorted(
            list(sampling_col)
        )


def test_cluster_sampling_num(grpc_stub):
    iris_path = os.path.join(dir_prefix, "../data/iris.csv")
    group_num = 1
    group_by = "Species"

    sampling_conf = SamplingConf(
        group_num=group_num,
        group_by=group_by,
        seed=1,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/iris_num_group_sampling")
    request = SamplingRequest(
        sampling_method=CLUSTER_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=iris_path,
        output_path=output_path,
        job_id="test_cluster_sampling_n",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")
    import pandas as pd

    pdf = pd.read_csv(response.data.sampled_path)
    assert int(pdf[[group_by]].drop_duplicates().count()) == group_num


def test_cluster_sampling_fraction(grpc_stub):
    iris_path = os.path.join(dir_prefix, "../data/iris.csv")

    sampling_conf = SamplingConf(
        fraction="0.5",
        group_by="Species",
        seed=1,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/iris_fraction_group_sampling")
    request = SamplingRequest(
        sampling_method=CLUSTER_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=iris_path,
        output_path=output_path,
        job_id="test_cluster_sampling_f",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")


def test_cluster_sampling_fraction_num(grpc_stub):
    iris_path = os.path.join(dir_prefix, "../data/iris.csv")
    group_num = 2
    group_by = "Species"

    sampling_conf = SamplingConf(
        fraction="0.5",
        group_by=group_by,
        group_num=group_num,
        seed=1,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/iris_fraction_num_group_sampling")
    request = SamplingRequest(
        sampling_method=CLUSTER_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=iris_path,
        output_path=output_path,
        job_id="test_cluster_sampling_fn",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")
    import pandas as pd

    pdf = pd.read_csv(response.data.sampled_path)
    assert int(pdf[[group_by]].drop_duplicates().count()) <= group_num


def test_cluster_sampling_empty(grpc_stub):
    iris_path = os.path.join(dir_prefix, "../data/empty.csv")
    group_num = 1
    group_by = "user"

    sampling_conf = SamplingConf(
        group_num=group_num,
        group_by=group_by,
        seed=1,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    output_path = os.path.join(dir_prefix, "./output/empty")
    request = SamplingRequest(
        sampling_method=CLUSTER_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=iris_path,
        output_path=output_path,
        job_id="test_cluster_sampling_empty",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")
    import pandas as pd

    pdf = pd.read_csv(response.data.sampled_path)
    assert int(pdf[[group_by]].drop_duplicates().count()) == 0
    for hook_msg in response.data.hook_msg:
        if hook_msg.hook_name == "EmptyDetectHook":
            for meta in hook_msg.meta:
                if meta.key == "is_not_empty":
                    assert meta.value == "False"


def test_generate_path(grpc_stub):
    sampling_conf = SamplingConf(
        fraction="0.2",
        with_replacement=False,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")

    request = SamplingRequest(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        job_id="test_random_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")


def test_simple_sampling(grpc_stub):
    sampling_col = ["# id", "X_0", "X_1", "X_2", "y"]
    sampling_conf = SamplingConf(count=1000, sampling_col=sampling_col)
    format_conf = FileFormatConf(with_header=True, sep=",")
    output_path = os.path.join(dir_prefix, "./output/test_simple_sampling")

    request = SamplingRequest(
        sampling_method=SIMPLE_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_simple_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")

    import pandas as pd

    df = pd.read_csv(response.data.sampled_path, sep=",")
    assert df.shape[0] == 1000
    if os.path.exists(response.data.sampled_path):
        assert sorted(list(df.columns)) == sorted(list(sampling_col))

    # bad count in simple sampling
    sampling_conf = SamplingConf(
        count=0,
    )
    request = SamplingRequest(
        sampling_method=SIMPLE_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_simple_sampling",
    )
    response = grpc_stub.SamplingJob(request)
    assert response.code == BAD_PARAM_ERROR


def test_stratified_sampling(grpc_stub):
    sampling_col = ["# id", "X_0", "X_1", "X_2", "y"]
    fraction = {"0": 0.3, "1": 0.7}
    import json

    sampling_conf = SamplingConf(
        fraction=json.dumps(fraction),
        with_replacement=True,
        stratified_key="y",
        sampling_col=sampling_col,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")
    output_path = os.path.join(dir_prefix, "./output/test_stratified_sampling")

    request = SamplingRequest(
        sampling_method=STRATIFIED_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_stratified_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")

    import pandas as pd

    df = pd.read_csv(response.data.sampled_path, sep=",")
    assert df[df["y"] == 1].shape[0] > (df[df["y"] == 0].shape[0]) * 2
    if os.path.exists(response.data.sampled_path):
        assert sorted(list(df.columns)) == sorted(list(sampling_col))


def test_simply_stratified_sampling(grpc_stub):
    fraction = "0.5"

    sampling_conf = SamplingConf(fraction=fraction, with_replacement=True, stratified_key="y")
    format_conf = FileFormatConf(with_header=True, sep=",")
    output_path = os.path.join(dir_prefix, "./output/test_simply_stratified_sampling")

    request = SamplingRequest(
        sampling_method=STRATIFIED_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_simply_stratified_sampling",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")

    import pandas as pd

    df = pd.read_csv(response.data.sampled_path, sep=",", index_col="# id")
    assert df[df["y"] == 1].shape[0] - (df[df["y"] == 0].shape[0]) < 500
    assert 24000 < df[df["y"] == 1].shape[0] < 26000
    assert 24000 < df[df["y"] == 0].shape[0] < 26000


def test_unknown_sampling_failure(grpc_stub):
    sampling_conf = SamplingConf()
    format_conf = FileFormatConf()
    output_path = os.path.join(dir_prefix, "./output/test_unknown_sampling_failure")

    request = SamplingRequest(
        sampling_method=UNKNOWN_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_unknown_sampling_failure",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == BAD_PARAM_ERROR


def test_unknown_type_failure(grpc_stub):
    sampling_conf = SamplingConf()
    format_conf = FileFormatConf()
    output_path = os.path.join(dir_prefix, "./output/test_unknown_type_failure")

    request = SamplingRequest(
        sampling_method=RANDOM_SAMPLING_METHOD,
        file_format=UNKNOWN_FORMAT,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_unknown_type_failure",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == BAD_PARAM_ERROR


def test_cancel(grpc_stub):
    request = CancelRequest(job_id="test_cancel")
    response = grpc_stub.CancelJob(request)
    return response.code == 0


def example_test_cancel(grpc_stub):
    sampling_conf = SamplingConf(
        fraction="0.2",
        with_replacement=False,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")
    output_path = os.path.join(dir_prefix, "./output/test_random_sampling")

    def cancel_job():
        request = CancelRequest(job_id="test_cancel")
        response = grpc_stub.CancelJob(request)
        return response.code == 0

    def start_job():
        request = SamplingRequest(
            sampling_method=RANDOM_SAMPLING_METHOD,
            file_format=FILE_FORMAT_CSV,
            sampling_conf=sampling_conf,
            format_conf=format_conf,
            input_path=input_path,
            output_path=output_path,
            job_id=f"test_cancel",
        )
        print(f"test_cancel: request for test_cancel")
        response = grpc_stub.SamplingJob(request)
        print(f"test_cancel: finish processing for test_cancel, response.code: {response.code}")
        return response.code == PROCESS_ERROR

    threads = []
    threads.append(JobThread(target=start_job))
    threads.append(JobThread(target=cancel_job))
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    results = [t.get_result() for t in threads]
    assert reduce(lambda x, y: x and y, results)


if __name__ == "__main__":
    pytest.main()



import os
from functools import reduce

import pytest
from utilities import JobThread

from sparksampling.error import EXHAUSTED_ERROR
from sparksampling.proto.sampling_service_pb2 import (
    FILE_FORMAT_CSV,
    RANDOM_SAMPLING_METHOD,
    STRATIFIED_SAMPLING_METHOD,
    FileFormatConf,
    SamplingConf,
    SamplingRequest,
)

dir_prefix = os.path.abspath(os.path.dirname(__file__))
input_path = os.path.join(dir_prefix, "../data/unbalance_500v50_10.csv")


@pytest.fixture
def exhausted_worker():
    from sparksampling.engine_factory import EngineFactory

    EngineFactory.register_all_engine()
    for engine in EngineFactory.engine_cls:
        engine.guarantee_worker = 0
    yield
    for engine in EngineFactory.engine_cls:
        engine.guarantee_worker = 10


def test_worker_limit(grpc_stub, exhausted_worker):
    # 本来想多线程测试，但是pytest似乎只能线性执行，所以上面把guarantee_worker设置为0，一定会触发EXHAUSTED_ERROR
    sampling_conf = SamplingConf(
        fraction="0.2",
        with_replacement=False,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")
    dir_prefix = os.path.abspath(os.path.dirname(__file__))
    input_path = os.path.join(dir_prefix, "../data/10w_x_10.csv")
    output_path = os.path.join(dir_prefix, "./output/test_random_sampling")

    def start_job(job_id):
        request = SamplingRequest(
            sampling_method=RANDOM_SAMPLING_METHOD,
            file_format=FILE_FORMAT_CSV,
            sampling_conf=sampling_conf,
            format_conf=format_conf,
            input_path=input_path,
            output_path=output_path,
            job_id=f"test_worker_limit-{job_id}",
        )
        print(f"test_worker_limit: request for job-{job_id}")
        response = grpc_stub.SamplingJob(request)
        print(
            f"test_worker_limit: finish processing for job-{job_id}, response.code: {response.code}"
        )
        return response.code == EXHAUSTED_ERROR

    threads = []
    for i in range(5):
        t = JobThread(target=start_job, kwargs={"job_id": i})
        threads.append(t)
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    results = [t.get_result() for t in threads]
    assert reduce(lambda x, y: x or y, results)


def test_stratified_sampling_reserve(grpc_stub):
    fraction = "0.5"

    sampling_conf = SamplingConf(
        fraction=fraction,
        with_replacement=True,
        stratified_key="# id",
        ensure_col=True,
    )
    format_conf = FileFormatConf(with_header=True, sep=",")
    output_path = os.path.join(dir_prefix, "./output/test_simply_stratified_sampling_reserve")

    request = SamplingRequest(
        sampling_method=STRATIFIED_SAMPLING_METHOD,
        file_format=FILE_FORMAT_CSV,
        sampling_conf=sampling_conf,
        format_conf=format_conf,
        input_path=input_path,
        output_path=output_path,
        job_id="test_simply_stratified_sampling_feature",
    )

    response = grpc_stub.SamplingJob(request)
    assert response.code == 0
    assert response.data.sampled_path.endswith(".csv")

    import pandas as pd

    df = pd.read_csv(response.data.sampled_path, sep=",", index_col="# id")
    assert df.shape[0] == 550


if __name__ == "__main__":
    pytest.main()
